-- Raven is an addon to monitor auras and cooldowns, providing timer bars, action bar highlights, and helpful notifications.

-- Bars.lua supports mapping auras to bars and grouping bars into multiple moveable frames using LibBars-1.0.
-- It has special case code for tooltips, test bars, and Death Knight runes.
-- There are no exported functions at this time other than those called to initialize and update bars.

local MOD = Raven
local L = LibStub("AceLocale-3.0"):GetLocale("Raven")
local LSPELL = MOD.LocalSpellNames
local LBF = LibStub("LibButtonFacade", true)
local media = LibStub("LibSharedMedia-3.0")
local tc = { r = 1, g = 1, b = 1, a = 1 }
local fishSpell = GetSpellInfo(7620)
local hidden = false
local detectedBar = { enableBar = true, sorder = 0 }
local settingsTemplate = {} -- settings are initialized from default bar group template
local defaultNotificationIcon = "Interface\\Icons\\Spell_Nature_WispSplode"
local testColors = { "Blue1", "Cyan", "Green1", "Yellow1", "Orange1", "Red1", "Pink", "Purple1", "Brown1", "Gray" }

-- Saved variables don't handle being set to nil properly so need to use alternate value to indicate an option has been turned off
local Off = 0 -- value used to designate an option is turned off
local function IsOff(value) return value == nil or value == Off end -- return true if option is turned off
local function IsOn(value) return value ~= nil and value ~= Off end -- return true if option is turned on

local classNames = { DEATHKNIGHT = "Death Knight", DRUID = "Druid", HUNTER = "Hunter", MAGE = "Mage", PALADIN = "Paladin",
	PRIEST = "Priest", ROGUE = "Rogue", SHAMAN = "Shaman", WARLOCK = "Warlock", WARRIOR = "Warrior" }
	
local colorTemplate = { timeColor = 0, iconColor = 0, labelColor = 0 }
local defaultWhite = { r = 1, g = 1, b = 1, a = 1 }
local function DefaultColor(c) return not c or not next(c) or ((c.r == 1) and (c.g == 1) and (c.b == 1) and (c.a == 1)) end

MOD.BarGroupTemplate = { -- default bar group settings
	enabled = true, locked = true, merged = false, linkSettings = false, checkCondition = false, noMouse = false, iconMouse = true,
	barColors = "Spell", bgColors = "Normal", iconColors = "None", combatTips = true, anchorTips = "DEFAULT", strata = "MEDIUM",
	useDefaultDimensions = true, useDefaultFontsAndTextures = true, useDefaultColors = true, sor = "A", reverseSort = false,
	anchor = false, anchorX = 0, anchorY = 0, anchorLastBar = false, anchorEmpty = false, configuration = 1, growDirection = true,
	pulseStart = false, pulseEnd = false, flashExpiring = false, flashTime = 5, hide = false, fade = false, delayTime = 5, fadeAlpha = 1,
	wrap = 16, wrapDirection = false, snapCenter = false, maxBars = 0, labelOffset = 0, labelInset = 0, labelWrap = false, labelAlign = "MIDDLE",
	timeOffset = 0, timeInset = 0, timeAlign = "normal", iconOffset = 0, iconInset = 0,
	showSolo = true, showParty = true, showRaid = true, showCombat = true, showOOC = true, showFishing = true,
	showResting = true, showMounted = true, showBlizz = true, showNotBlizz = true,
	detectBuffs = false, detectDebuffs = false, detectDispellable = false, detectStealable = false,
	detectCooldowns = false, detectBuffsMonitor = "player", detectBuffsCastBy = "player", detectDebuffsMonitor = "player",
	detectDebuffsCastBy = "player", detectCooldownsBy = "player", detectTracking = false, detectOnlyTracking = false,
	detectSpellCooldowns = true, detectTrinketCooldowns = true, detectPotionCooldowns = true, detectOtherCooldowns = true,
	detectSharedFrostTraps = true, detectSharedFireTraps = true, detectSharedShocks = true, detectSharedCrusader = true, detectRuneCooldowns = false,
	setDuration = false, uniformDuration = 120, checkDuration = false, minimumDuration = true, filterDuration = 120,
	checkTimeLeft = false, minimumTimeLeft = true, filterTimeLeft = 120, showNoDuration = false, showOnlyNoDuration = false,
	showNoDurationBackground = false, timeFormat = 6, timeSpaces = false, timeCase = false,
	filterBuff = true, filterBuffLink = true, filterDebuff = true, filterDebuffLink = true, filterCooldown = true, filterCooldownLink = true,
	showBuff = false, showDebuff = false, showCooldown = false, filterBuffBars = false, filterDebuffBars = false, filterCooldownBars = false,
}

MOD.BarGroupLayoutTemplate = { -- all the bar group settings involved in layout configuration
	barWidth = 0, barHeight = 0, iconSize = 0, scale = 0, spacingX = 0, spacingY = 0, iconOffsetX = 0, iconOffsetY = 0,
	useDefaultDimensions = 0, configuration = 0, growDirection = 0, wrap = 0, wrapDirection = 0, snapCenter = 0, maxBars = 0,
	labelOffset = 0, labelInset = 0, labelWrap = 0, labelAlign = 0, timeOffset = 0, timeInset = 0, timeAlign = 0, iconOffset = 0, iconInset = 0,
	hideIcon = 0, hideClock = 0, hideBar = 0, hideSpark = 0, hideValue = 0, hideLabel = 0, hideCount = 0, showTooltips = 0
}

-- Initialize bar groups from those specified in the profile after, for example, a reloadUI or reset profile
function MOD:InitializeBars()
	local bgs = Nest_GetBarGroups()
	if bgs then for _, bg in pairs(bgs) do Nest_DeleteBarGroup(bg) end end -- first remove any bar groups represented in the graphics library
		
	for _, bg in pairs(MOD.db.profile.BarGroups) do -- then set up the ones specified in the profile
		if IsOn(bg) then
			for n, k in pairs(MOD.db.global.Defaults) do if bg[n] == nil then bg[n] = k end end -- add default settings for layout, fonts and textures
			for n, k in pairs(MOD.BarGroupTemplate) do if bg[n] == nil then bg[n] = k end end -- add defaults from the bar group template
			for n in pairs(colorTemplate) do if bg[n] == nil then bg[n] = Raven_CopyColor(defaultWhite) end end -- default basic colors
			MOD:InitializeBarGroup(bg, 0, 0)
		end
	end
	if LBF then -- restore ButtonFacade skin information from the profile
		local lbf = MOD.db.profile.ButtonFacadeSkin -- table containing saved skin data
		if lbf then Nest_SetButtonFacadeData(lbf.SkinID, lbf.Gloss, lbf.Backdrop, lbf.Colors) end
	end
	MOD:UpdateAllBarGroups() -- this is done last to get all positions updated correctly
end

-- Finalize bar groups prior to logout, stripping out all values that match current defaults
function MOD:FinalizeBars()
	for bn, bg in pairs(MOD.db.profile.BarGroups) do
		if IsOn(bg) then
			bg.cache = nil -- delete bar group cache contents
			for n, k in pairs(MOD.db.global.Defaults) do if bg[n] == k then bg[n] = nil end end -- remove default settings for layout, fonts and textures
			for n, k in pairs(MOD.BarGroupTemplate) do if bg[n] == k then bg[n] = nil end end -- remove defaults from the bar group template
			for n in pairs(colorTemplate) do if DefaultColor(bg[n]) then bg[n] = nil end end -- detect basic colors set to defaults
		else
			MOD.db.profile.BarGroups[bn] = nil -- okay to delete these since no default bar groups
		end
	end
	if LBF then -- save ButtonFacade skin information in the profile
		local lbf = MOD.db.profile.ButtonFacadeSkin -- table containing saved skin data
		if not lbf then lbf = {}; MOD.db.profile.ButtonFacadeSkin = lbf end
		lbf.SkinID, lbf.Gloss, lbf.Backdrop, lbf.Colors = Nest_GetButtonFacadeData()
	end
end

-- Remove default values from bar group settings
function MOD:FinalizeSettings()
	for _, settings in pairs(MOD.db.global.Settings) do
		for n, k in pairs(settingsTemplate) do if settings[n] == k then settings[n] = nil end end -- remove values still set to defaults
		for n in pairs(colorTemplate) do if DefaultColor(settings[n]) then settings[n] = nil end end -- detect basic colors set to defaults
	end
end

-- Raven is disabled so hide all features
function MOD:HideBars()
	if not hidden then
		for _, bp in pairs(MOD.db.profile.BarGroups) do
			if IsOn(bp) then MOD:ReleaseBarGroup(bp) end
		end
		hidden = true
	end
end

-- Initialize bar group settings by adding default values if necessary
function MOD:InitializeSettings()
	for n, k in pairs(MOD.BarGroupTemplate) do settingsTemplate[n] = k end -- initialize the settings template from bar group defaults
	for n, k in pairs(MOD.db.global.Defaults) do settingsTemplate[n] = k end -- add default settings for layout-fonts-textures
	settingsTemplate.enabled = nil; settingsTemplate.locked = nil; settingsTemplate.merged = nil; settingsTemplate.linkSettings = nil	
	for _, settings in pairs(MOD.db.global.Settings) do
		for n, k in pairs(settingsTemplate) do if settings[n] == nil then settings[n] = k end end -- add missing defaults from settings template
		for n in pairs(colorTemplate) do if settings[n] == nil then settings[n] = Raven_CopyColor(defaultWhite) end end -- default basic colors
	end
end

-- Fire off test bars for this bar group, remove if any already exist
local function TestBarGroup(bp)
	local bg = Nest_GetBarGroup(bp.name)
	if bg then
		local found = false
		local icon = "Interface\\Icons\\INV_Drink_20"
		for i = 1, 10 do
			local bar = Nest_GetBar(bg, ">>Test<<" .. string.format("%02d", i))
			if bar then found = true; Nest_DeleteBar(bg, bar) end
		end
		if not found then
			for i = 1, 10 do			
				local bar = Nest_CreateBar(bg, ">>Test<<" .. string.format("%02d", i))
				if bar then
					local c = MOD.ColorPalette[testColors[i]]
					Nest_SetColors(bar, c.r, c.g, c.b, 1, c.r, c.g, c.b, 1, c.r, c.g, c.b, 1)
					Nest_SetLabel(bar, L["Test Bar"] .. " " .. i)
					Nest_SetIcon(bar, icon)
					Nest_SetCount(bar, i)
					Nest_StartTimer(bar, i*5, 60, 60)
					Nest_SetAttribute(bar, "updated", true)
				end
			end
		end
	end
end

-- Make sure not to delete any unexpired test bars
local function UpdateTestBars(bg)
	for i = 1, 10 do
		local bar = Nest_GetBar(bg, ">>Test<<" .. string.format("%02d", i))
		if bar then
			local timeLeft = Nest_GetTimes(bar)
			if timeLeft and (timeLeft > 0) then Nest_SetAttribute(bar, "updated", true) end
		end
	end
end

-- Show tooltip when entering a bar group anchor
local function Anchor_OnEnter(anchor, bgName)
	if GetCVar("UberTooltips") == "1" then
		GameTooltip_SetDefaultAnchor(GameTooltip, anchor)
	else
		GameTooltip:SetOwner(anchor, "ANCHOR_BOTTOMLEFT")
	end
	local bg, bgType, attachment = Nest_GetBarGroup(bgName), L["Custom Bar Group"], nil
	if bg then
		if Nest_GetBarGroupAttribute(bg, "isAuto") then bgType = L["Auto Bar Group"] end
		attachment = Nest_GetBarGroupAttribute(bg, "attachment")
	end
	GameTooltip:AddDoubleLine("Raven", bgType)
	if attachment then
		GameTooltip:AddLine(L["Anchor attached"] .. attachment .. '"')
		GameTooltip:AddLine(L["Anchor left click 1"])
	else
		GameTooltip:AddLine(L["Anchor left click 2"])
	end
	GameTooltip:AddLine(L["Anchor right click"])
	GameTooltip:AddLine(L["Anchor shift left click"])
	GameTooltip:AddLine(L["Anchor shift right click"])
	GameTooltip:AddLine(L["Anchor alt left click"])
	GameTooltip:AddLine(L["Anchor alt right click"])
	GameTooltip:Show()
end

-- Hide tooltip when leaving a bar group anchor
local function Anchor_OnLeave(anchor, bgName)
	GameTooltip:Hide()
end

-- Callback function for tracking location of the bar group
local function Anchor_Moved(anchor, bgName)
	local bp = MOD.db.profile.BarGroups[bgName]
	if IsOn(bp) then
		local bg = Nest_GetBarGroup(bgName)
		if bg then
			bp.pointX, bp.pointXR, bp.pointY, bp.pointYT, bp.pointW, bp.pointH = Nest_GetAnchorPoint(bg) -- returns fractions from display edge
			if bp.anchor then bp.anchor = false end -- no longer anchored to other bar groups			
			if bp.linkSettings then
				local settings = MOD.db.global.Settings[bp.name] -- when updating a bar group with linked settings always overwrite position
				if settings then
					settings.pointX = bp.pointX; settings.pointXR = bp.pointXR; settings.pointY = bp.pointY; settings.pointYT = bp.pointYT
					settings.pointW = bp.pointW; settings.pointH = bp.pointH; settings.anchor = bp.anchor
				end
			end
			Anchor_OnLeave(anchor) -- turn off tooltip while moving the anchor
			MOD.updateOptions = true -- if options panel is open then update it in case viewing position info
		end
		return
	end
end

-- Callback function for when a bar group anchor is clicked with a modifier key down
-- Shift left click is test bars, right click is "toggle lock and hide",
local function Anchor_Clicked(anchor, bgName, button)
	local shiftLeftClick = (button == "LeftButton") and (IsShiftKeyDown() == 1)
	local shiftRightClick = (button == "RightButton") and (IsShiftKeyDown() == 1)
	local altLeftClick = (button == "LeftButton") and (IsAltKeyDown() == 1)
	local altRightClick = (button == "RightButton") and (IsAltKeyDown() == 1)
	local rightClick = (button == "RightButton")

	local bp = MOD.db.profile.BarGroups[bgName]
	if IsOn(bp) then
		if shiftLeftClick then -- test bars
			TestBarGroup(bp)
		elseif shiftRightClick then -- toggle grow up/down
			bp.growDirection = not bp.growDirection
		elseif altLeftClick then -- toggle options menu
			MOD:OptionsPanel()
		elseif altRightClick then -- cycle through configurations
			if bp.configuration > Nest_MaxBarConfiguration then -- special case order for cycling icon configurations
				local c, i = bp.configuration, Nest_MaxBarConfiguration + 1
				if c == i then c = i + 2 elseif c == (i + 1) then c = i + 3 elseif c == (i + 2) then c = i + 1 else c = i end
				bp.configuration = c
			else
				bp.configuration = bp.configuration + 1
				if bp.configuration == (Nest_MaxBarConfiguration + 1) then bp.configuration = 1 end
			end
		elseif rightClick then -- lock and hide
			bp.locked = true
		end
		MOD:UpdateBarGroup(bp)
		MOD.updateOptions = true -- if options panel is open then update it in case viewing configuration info
		return
	end
end

-- Update linked settings. If dir is true then update the shared settings, otherwise update the bar group.
-- Also, if dir is true, create a linked settings table if one doesn't yet exist.
local function UpdateLinkedSettings(bp, dir)
	local settings = MOD.db.global.Settings[bp.name]
	if not settings then
		if not dir then return end
		settings = {}
		MOD.db.global.Settings[bp.name] = settings
	end

	local p, q = settings, bp
	if dir then p = bp; q = settings end

	for n in pairs(settingsTemplate) do q[n] = p[n] end -- copy every setting in the template
	q.pointX = p.pointX; q.pointXR = p.pointXR; q.pointY = p.pointY; q.pointYT = p.pointYT -- always copy the location
	q.pointW = p.pointW; q.pointH = p.pointH
end

-- Update a linked filter list. If dir is true then update the shared list, otherwise update the bar group's list.
-- Also, if dir is true, create a linked filter list if one doesn't yet exist.
local function UpdateLinkedFilter(bp, dir, filterType)
	local shared = MOD.db.global["Filter" .. filterType]
	local bgname = "filter" ..  filterType .. "List"

	if not shared[bp.name] then
		if not dir or not bp[bgname] or not next(bp[bgname], nil) then return end
		shared[bp.name] = {}
	end
	
	if not bp[bgname] then bp[bgname] = {} end

	local p, q = shared[bp.name], bp[bgname]
	if dir then p = bp[bgname]; q = shared[bp.name] end

	for _, v in pairs(q) do if not p[v] then q[v] = nil end end -- delete any keys in q not in p
	for _, v in pairs(p) do if not q[v] then q[v] = v end end -- copy everything from p to q
end

-- Initialize a bar group from a shared filter list, if any.
-- This function is called whenever filterLink is changed.
function MOD:InitializeFilterList(bp, filterType)
	if bp and bp["filter" .. filterType] and bp["filter" .. filterType .. "Link"] then
		UpdateLinkedFilter(bp, false, filterType) -- use the shared settings, if any, for a linked layout
	end
end

-- Get spell associated with a bar
function MOD:GetAssociatedSpellForBar(bar)
	if bar.barType == "Notification" then
		local sp = nil
		if bar.action then sp = MOD:GetConditionSpell(bar.action) end
		return sp
	end
	return bar.action
end

-- Get icon for the spell associated with a bar, returns nil if none found
function MOD:GetIconForBar(bar)
	local sp = MOD:GetAssociatedSpellForBar(bar)
	if sp then return MOD:GetIcon(sp) end
	return nil
end

-- Get color for the spell associated with a bar, returns nil if none found
function MOD:GetSpellColorForBar(bar)
	local c = bar.color -- get override if one is set
	if not c then
		local sp = MOD:GetAssociatedSpellForBar(bar)
		if sp then c = MOD:GetColor(sp) end
	end
	return c
end

-- Set the bar's current spell color using an override if not linked (note bar.colorLink uses inverted value from expected)
function MOD:SetSpellColorForBar(bar, r, g, b, a)
	local c = bar.color -- check if using an override
	if c then c.r = r; c.g = g; c.b = b; c.a = a; return end
	if not bar.colorLink then -- set the shared color for bars with the same associated spell
		local sp = MOD:GetAssociatedSpellForBar(bar)
		if sp then c = MOD:GetColor(sp) end
		if c then c.r = r; c.g = g; c.b = b; c.a = a; return end
		c = { r = r, g = g, b = b, a = a }
		if sp then MOD:SetColor(sp, c) else bar.color = c end
	else
		bar.color = { r = r, g = g, b = b, a = a } -- create an override
	end
end

-- Either link or decouple the bar's color from the color cache for it's associated spell
function MOD:LinkSpellColorForBar(bar)
	local c = MOD:GetSpellColorForBar(bar)
	if not bar.colorLink then -- link to the color cache, copying current setting, if any, to the color cache
		if c and bar.color then local d = bar.color; c.r = d.r; c.g = d.g; c.b = d.b; c.a = d.a end
		bar.color = nil -- delete override to revert back to color cache or default for bar type
	else -- decouple from the color cached, copying current setting, if any, to a new override
		if c then bar.color = { r = c.r, g = c.g, b = c.b, a = c.a } end
	end
end

-- Get the right color for the bar based on bar group settings
local function GetColorForBar(bg, bar, btype)
	local bt, ba, b, c = bar.barType, bar.action, nil, nil
	if bg.barColors == "Class" then
		if bt == "Buff" then b = MOD.BuffTable[ba] end
		if bt == "Debuff" then b = MOD.DebuffTable[ba] end
		if bt == "Cooldown" then b = MOD.CooldownTable[ba] end
		if b and b.class then
			if CUSTOM_CLASS_COLORS then c = CUSTOM_CLASS_COLORS[b.class] else c = RAID_CLASS_COLORS[b.class] end
		end
	elseif bg.barColors == "Spell" then
		if bar.color then c = bar.color else c = MOD:GetSpellColorForBar(bar) end
	elseif bg.barColors == "Custom" then
		c = bg.fgColor or defaultWhite
	end
	if not c then -- get the best default color for this bar type
		local cc = not bg.useDefaultColors -- indicates the bar group has overrides for standard colors
		c = cc and bg.buffColor or MOD.db.global.DefaultBuffColor -- use this as default in case unrecognized bar type
		if bt == "Debuff" then
			c = cc and bg.debuffColor or MOD.db.global.DefaultDebuffColor
			if btype then
				if btype == "Poison" then c = cc and bg.poisonColor or MOD.db.global.DefaultPoisonColor end
				if btype == "Curse" then c = cc and bg.curseColor or MOD.db.global.DefaultCurseColor end
				if btype == "Magic" then c = cc and bg.magicColor or MOD.db.global.DefaultMagicColor end
				if btype == "Disease" then c = cc and bg.diseaseColor or MOD.db.global.DefaultDiseaseColor end
			end
		end
		if bt == "Cooldown" then c = cc and bg.cooldownColor or MOD.db.global.DefaultCooldownColor end
		if bt == "Notification" then c = cc and bg.notificationColor or MOD.db.global.DefaultNotificationColor end
	end
	c.a = 1 -- always set alpha to 1 for bar colors
	return c
end

-- Get the standard debuff color for the bar, return nil if not a standard type of debuff
function MOD:GetDebuffColorForBar(bg, bar, btype)
	local cc = not bg.useDefaultColors -- indicates the bar group has overrides for standard colors
	local c = cc and bg.debuffColor or MOD.db.global.DefaultDebuffColor
	if bar.barType == "Debuff" then
		if btype then
			if btype == "Poison" then c = cc and bg.poisonColor or MOD.db.global.DefaultPoisonColor end
			if btype == "Curse" then c = cc and bg.curseColor or MOD.db.global.DefaultCurseColor end
			if btype == "Magic" then c = cc and bg.magicColor or MOD.db.global.DefaultMagicColor end
			if btype == "Disease" then c = cc and bg.diseaseColor or MOD.db.global.DefaultDiseaseColor end
		end
	end
	return c
end

-- Initialize bar group in LibBars and set default values from those set in profile
function MOD:InitializeBarGroup(bp, offsetX, offsetY)
	local bg = Nest_GetBarGroup(bp.name)
	if not bg then bg = Nest_CreateBarGroup(bp.name) end
	if bp.linkSettings then UpdateLinkedSettings(bp, false) end
	if bp.auto then -- initialize the auto bar group filter lists
		if (bp.filterBuff or bp.showBuff) and bp.filterBuffLink then UpdateLinkedFilter(bp, false, "Buff") end -- shared settings for buffs
		if (bp.filterDebuff or bp.showDebuff) and bp.filterDebuffLink then UpdateLinkedFilter(bp, false, "Debuff") end -- shared settings for debuffs
		if (bp.filterCooldown or bp.showCooldown) and bp.filterCooldownLink then UpdateLinkedFilter(bp, false, "Cooldown") end -- shared settings for buffs
	end
	if not bp.pointX or not bp.pointY then bp.pointX = 0.5 + (offsetX / 600); bp.pointXR = nil; bp.pointY = 0.5 + (offsetY / 600); bp.pointYT = nil end
	if not bp.pointW or not bp.pointH then bp.pointW = MOD.db.global.Defaults.barWidth; bp.pointH = MOD.db.global.Defaults.barHeight end
	MOD:SetBarGroupPosition(bp)
	Nest_SetBarGroupCallbacks(bg, Anchor_Moved, Anchor_Clicked, Anchor_OnEnter, Anchor_OnLeave)
end

-- Initialize a bar group from linked settings, if any, and always update the bar group location.
-- This function is called whenever linkSettings is changed.
function MOD:InitializeBarGroupSettings(bp)
	if bp and bp.enabled then
		if bp.linkSettings then UpdateLinkedSettings(bp, false) end
		MOD:SetBarGroupPosition(bp)
	end
end

-- Save bar group settings into the linked settings.
function MOD:SaveBarGroupSettings(bp) UpdateLinkedSettings(bp, true) end

-- Validate and update a bar group's display position. If linked, also update the position in the linked settings.
function MOD:SetBarGroupPosition(bp)
	local scale = bp.useDefaultDimensions and MOD.db.global.Defaults.scale or bp.scale or 1
	local bg = Nest_GetBarGroup(bp.name)
	if bg then Nest_SetAnchorPoint(bg, bp.pointX, bp.pointXR, bp.pointY, bp.pointYT, scale, bp.pointW, bp.pointH) end
	if bp.linkSettings then
		local settings = MOD.db.global.Settings[bp.name] -- when updating a bar group with linked settings always overwrite position
		if settings then
			settings.pointX = bp.pointX; settings.pointXR = bp.pointXR; settings.pointY = bp.pointY; settings.pointYT = bp.pointYT
			settings.pointW = bp.pointW; settings.pointH = bp.pointH
		end
	end
end

-- Returns whether solo, party or raid
local function PartyInfo()
	if GetNumRaidMembers() > 0 then return "raid" end
	if GetNumPartyMembers() > 0 then return "party" end
	return "solo"
end

-- Set an entry in a bar group cache block
local function SetCache(bg, block, name, value)
	if not bg.cache then bg.cache = {} end
	if not bg.cache.block then bg.cache.block = {} end
	bg.cache.block[name] = value
end

-- Get a value from a bar group cache block
local function GetCache(bg, block, name)
	if not bg.cache or not bg.cache.block then return nil end
	return bg.cache.block[name]
end

-- Reset a bar group cache block
local function ResetCache(bg, block)
	if bg.cache and bg.cache.block then table.wipe(bg.cache.block) end
end

-- Bar sorting functions, assumes sort order was built into name
local function SortAlphaUp(a, b) return a.name > b.name end
local function SortAlphaDown(a, b) return a.name < b.name end
local function SortTimeLeftUp(a, b) if a.value ~= b.value then return a.value > b.value else return a.name > b.name end end
local function SortTimeLeftDown(a, b) if a.value ~= b.value then return a.value < b.value else return a.name < b.name end end
local function SortDurationUp(a, b) if a.maxValue ~= b.maxValue then return a.maxValue > b.maxValue else return a.name > b.name end end
local function SortDurationDown(a, b) if a.maxValue ~= b.maxValue then return a.maxValue < b.maxValue else return a.name < b.name end end

-- Update a bar group in LibBars with the current values in the profile
function MOD:UpdateBarGroup(bp)
	if bp.enabled then
		if bp.linkSettings then UpdateLinkedSettings(bp, true) end -- update shared settings in a linked bar group
		if bp.auto then -- update auto bar group filter lists
			if (bp.filterBuff or bp.showBuff) and bp.filterBuffLink then UpdateLinkedFilter(bp, true, "Buff") end -- shared settings for buffs
			if (bp.filterDebuff or bp.showDebuff) and bp.filterDebuffLink then UpdateLinkedFilter(bp, true, "Debuff") end -- shared settings for debuffs
			if (bp.filterCooldown or bp.showCooldown) and bp.filterCooldownLink then UpdateLinkedFilter(bp, true, "Cooldown") end -- shared settings for buffs
		end

		ResetCache(bp, "Buff"); ResetCache(bp, "Debuff"); ResetCache(bp, "Cooldown")
		if bp.bars then -- create caches for buff, debuff, cooldown actions
			for _, b in pairs(bp.bars) do
				local t = b.barType
				if (t == "Buff") or (t == "Debuff") then
					SetCache(bp, t, b.action, b.monitor)
				elseif t == "Cooldown" then
					SetCache(bp, t, b.action, true)
				end
			end
		end

		local bg = Nest_GetBarGroup(bp.name)
		if not bg then MOD:InitializeBarGroup(bp); bg = Nest_GetBarGroup(bp.name) end

		if bp.useDefaultDimensions then MOD:CopyDimensions(MOD.db.global.Defaults, bp) end
		if bp.useDefaultFontsAndTextures then MOD:CopyFontsAndTextures(MOD.db.global.Defaults, bp) end		
		local fgtexture = media:Fetch("statusbar", bp.texture)
		local bgtexture = fgtexture
		if bp.bgtexture then bgtexture = media:Fetch("statusbar", bp.bgtexture) end
		Nest_SetBarGroupLock(bg, bp.locked)
		Nest_SetBarGroupBarLayout(bg, bp.barWidth, bp.barHeight, bp.iconSize, bp.scale, bp.spacingX, bp.spacingY, bp.iconOffsetX, bp.iconOffsetY,
			bp.labelOffset, bp.labelInset, bp.labelWrap, bp.labelAlign, bp.timeOffset, bp.timeInset, bp.timeAlign, bp.iconOffset, bp.iconInset,
			bp.configuration, bp.growDirection, bp.wrap, bp.wrapDirection, bp.snapCenter, bp.maxBars, bp.strata)
		Nest_SetBarGroupLabelFont(bg, media:Fetch("font", bp.labelFont), bp.labelFSize, bp.labelAlpha, bp.labelColor, bp.labelOutline)
		Nest_SetBarGroupTimeFont(bg, media:Fetch("font", bp.timeFont), bp.timeFSize, bp.timeAlpha, bp.timeColor, bp.timeOutline)
		Nest_SetBarGroupIconFont(bg, media:Fetch("font", bp.iconFont), bp.iconFSize, bp.iconAlpha, bp.iconColor, bp.iconOutline)
		Nest_SetBarGroupTextures(bg, fgtexture, bp.fgAlpha, bgtexture, bp.bgAlpha, not bp.showNoDurationBackground,
			bp.fgSaturation, bp.fgBrightness, bp.bgSaturation, bp.bgBrightness)
		Nest_SetBarGroupVisibles(bg, not bp.hideIcon, not bp.hideClock, not bp.hideBar, not bp.hideSpark, not bp.hideLabel, not bp.hideValue)
		Nest_SetBarGroupAttribute(bg, "pulseStart", bp.pulseStart) -- pulse icon at start
		Nest_SetBarGroupAttribute(bg, "pulseEnd", bp.pulseEnd) -- pulse icon when expiring
		Nest_SetBarGroupAttribute(bg, "noMouse", bp.noMouse) -- disable interactivity
		Nest_SetBarGroupAttribute(bg, "iconMouse", bp.iconMouse) -- mouse-only interactivity
		Nest_SetBarGroupAttribute(bg, "anchorTips", bp.anchorTips) -- manual tooltip anchor
		Nest_SetBarGroupAttribute(bg, "isAuto", bp.auto) -- save for tooltip
		Nest_SetBarGroupAttribute(bg, "attachment", bp.anchor) -- save for tooltip
		Nest_SetBarGroupTimeFormat(bg, bp.timeFormat, bp.timeSpaces, bp.timeCase)
		local sf = "alpha"; if bp.sor == "T" then sf = "time" elseif bp.sor == "D" then sf = "duration" elseif bp.sor == "S" then sf = "start" end
		Nest_BarGroupSortFunction(bg, sf, bp.reverseSort)
	else
		MOD:ReleaseBarGroup(bp)
	end
end

-- Update the positions of all anchored bar groups plus make sure valid positions in all bar groups
local function UpdatePositions()
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- update bar group positions including relative ones if anchored
		if IsOn(bp) then
			local bg = Nest_GetBarGroup(bp.name)
			if bg and bg.configuration then -- make sure already configured
				if not bp.pointX or not bp.pointY then -- if not valid then move to center
					bp.pointX = 0.5; bp.pointXR = nil; bp.pointY = 0.5; bp.pointYT = nil
					Nest_SetAnchorPoint(bg, bp.pointX, bp.pointXR, bp.pointY, bp.pointYT, bp.scale or 1, nil, nil)
				end	
				if bp.anchor then
					local abp = MOD.db.profile.BarGroups[bp.anchor]
					if IsOn(abp) and abp.enabled then -- make sure the anchor is actually around to attach
						Nest_SetRelativeAnchorPoint(bg, bp.anchor, bp.anchorX, bp.anchorY, bp.anchorLastBar, bp.anchorEmpty)
					end
				else
					Nest_SetRelativeAnchorPoint(bg, nil) -- reset the relative anchor point if none set
				end
				bp.pointX, bp.pointXR, bp.pointY, bp.pointYT, bp.pointW, bp.pointH = Nest_GetAnchorPoint(bg) -- returns fractions from display edge
			end
		end
	end
end

-- Update all the bar groups, this is necessary when changing stuff that can affect bars in multiple groups (e.g., buff colors and labels)
function MOD:UpdateAllBarGroups()
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- update for changed bar group settings
		if IsOn(bp) then MOD:UpdateBarGroup(bp) end
	end
end

-- Lock or unlock all bar groups
function MOD:LockBarGroups(lock)
	for _, bp in pairs(MOD.db.profile.BarGroups) do
		if IsOn(bp) then bp.locked = lock end
	end
	MOD:UpdateAllBarGroups()
end

-- Toggle locking of bar groups 
function MOD:ToggleBarGroupLocks()
	-- Look in the profile table to determine current state
	local anyLocked = false
	for _, bp in pairs(MOD.db.profile.BarGroups) do
		if IsOn(bp) and bp.locked then anyLocked = true break end
	end
	-- Now go back through and set all to same state (if any locked then unlock all)
	MOD:LockBarGroups(not anyLocked)
end

-- Release all the bars in the named bar group in LibBars
function MOD:ReleaseBarGroup(bp)
	if bp then 
		local bg = Nest_GetBarGroup(bp.name)
		if bg then Nest_DeleteBarGroup(bg) end
	end
end

-- Show tooltip when entering a bar
local function Bar_OnEnter(frame, bgName, barName, ttanchor)
	if (ttanchor == "DEFAULT") and (GetCVar("UberTooltips") == "1") then
		GameTooltip_SetDefaultAnchor(GameTooltip, frame)
	else
		if not ttanchor or (ttanchor == "DEFAULT") then ttanchor = "ANCHOR_BOTTOMLEFT" else ttanchor = "ANCHOR_" .. ttanchor end
		GameTooltip:SetOwner(frame, ttanchor)
	end
	local bg = Nest_GetBarGroup(bgName)
	if not bg then return end
	local bar = Nest_GetBar(bg, barName)
	if not bar then return end	
	local tt = Nest_GetAttribute(bar, "tooltipType")
	local id = Nest_GetAttribute(bar, "tooltipID")
	local unit = Nest_GetAttribute(bar, "tooltipUnit")	
	if tt == "text" then
		GameTooltip:SetText(id)
	elseif (tt == "inventory") or (tt == "weapon") then
		local slot = GetInventorySlotInfo(id)
		if slot then
			GameTooltip:SetInventoryItem("player", slot)
		end
	elseif (tt == "spell link") or (tt == "item link") then
		GameTooltip:SetHyperlink(id)
	elseif tt == "buff" then
		GameTooltip:SetUnitBuff(unit, id)
	elseif tt == "debuff" then
		GameTooltip:SetUnitDebuff(unit, id)
	elseif tt == "vehicle buff" then
		GameTooltip:SetUnitBuff("vehicle", id)
	elseif tt == "vehicle debuff" then
		GameTooltip:SetUnitDebuff("vehicle", id)
	elseif tt == "tracking" then
		GameTooltip:SetText(id) -- id is localized name of tracking type
	elseif tt == "spell" then
		GameTooltip:SetText(id)
	elseif tt == "notification" then
		GameTooltip:AddDoubleLine(id, "Notification")
		local ct = MOD.db.profile.Conditions[MOD.myClass]
		if ct then
			local c = ct[unit]
			if IsOn(c) and c.tooltip then GameTooltip:AddLine(MOD:GetConditionText(c.name), 1, 1, 1, true) end
		end
	end
	GameTooltip:Show()
end

-- Hide tooltip when leaving a bar
local function Bar_OnLeave(frame, bgName, barName, ttanchor)
	GameTooltip:Hide()
end

-- Handle left clicking on a tracking bar to show options
local function Bar_OnClick(frame, bgName, barName, button)
	local bg = Nest_GetBarGroup(bgName)
	if not bg then return end
	local bar = Nest_GetBar(bg, barName)
	if not bar then return end	
	local tt = Nest_GetAttribute(bar, "tooltipType")
	local unit = Nest_GetAttribute(bar, "tooltipUnit")
	if (button == "LeftButton") and (unit == "player") and (tt == "tracking") then
		if GameTooltip:GetOwner() == frame then GameTooltip:Hide() end
		ToggleDropDownMenu(1, nil, MiniMapTrackingDropDown, frame, 0, 0)
	end
end

-- Return true if time and duration pass a bar group's filters
local function CheckTimeAndDuration(bp, timeLeft, duration)
	if (timeLeft == 0) and (duration == 0) then -- test for unlimited duration
		if not bp.showNoDuration then return false end
	else
		if bp.showNoDuration and bp.showOnlyNoDuration then return false end
		if bp.checkDuration and bp.filterDuration then
			if bp.minimumDuration then if duration < bp.filterDuration then return false end
			elseif duration >= bp.filterDuration then return false end
		end
		if bp.checkTimeLeft and bp.filterTimeLeft then
			if bp.minimumTimeLeft then if timeLeft < bp.filterTimeLeft then return false end
			elseif timeLeft >= bp.filterTimeLeft then return false end
		end
	end
	return true
end

-- Manage a bar, creating one if not currently active, otherwise updating as necessary
-- Use originating bar group (bp) for filtering, display bar group (vbp) for appearance options
local function UpdateBar(bp, vbp, bg, b, icon, timeLeft, duration, count, btype, ttType, ttID, ttUnit)
	if duration > 0 then -- check if timer bar
		local elapsed = duration - timeLeft
		if (b.hide and (elapsed > (b.delayTime or 5))) or (bp.hide and (elapsed > (bp.delayTime or 5))) then return end
	end
	
	local bar, barname, label, src = nil, b.barLabel .. b.uniqueID, b.barLabel, b.barSource
	if vbp.sor == "C" then -- prefixes are added to the barname to facilitate sorting
		if not src or (src == "Racial") or (src == "Spell") or (src == "Detected") then src = "ZZZZ" end
		barname = "ZZZZ" .. string.upper(src) .. barname
	end
	if vbp.sor == "X" then barname = string.format("%05d ", b.sorder) .. barname end
	
	local maxTime = duration
	if vbp.setDuration then maxTime = vbp.uniformDuration end -- override with uniform duration for all bars in group

	local c = GetColorForBar(vbp, b, btype)
	local iconCount = nil
	if count and (count > 1) then
		if not vbp.hideCount then label = string.format("%s (%d)", label, count) end
		if not vbp.hideIcon then iconCount = count end
	end
	
	bar = Nest_GetBar(bg, barname)
	if not (bp.showNoDuration and bp.showOnlyNoDuration) and not ((timeLeft == 0) and (duration == 0)) then -- bar with duration
		if bar and Nest_IsTimer(bar) then -- existing timer bar
			local oldTimeLeft, oldDuration, oldMaxTime = Nest_GetTimes(bar)
			if (duration ~= oldDuration) or maxTime ~= oldMaxTime or (math.abs(timeLeft - oldTimeLeft) > 0.5) then
				Nest_StartTimer(bar, timeLeft, duration, maxTime) -- update if the bar is out of sync
			end
		else
			if bar then Nest_DeleteBar(bg, bar); bar = nil end
			bar = Nest_CreateBar(bg, barname)
			if bar then Nest_StartTimer(bar, timeLeft, duration, maxTime) end
		end
	elseif bp.showNoDuration or (b.barType == "Notification") or b.enableReady then -- bars without duration
		if bar and Nest_IsTimer(bar) then Nest_DeleteBar(bg, bar); bar = nil end
		if not bar then
			bar = Nest_CreateBar(bg, barname)
			if bar and b.barType == "Notification" then Nest_SetFlash(bar, b.flash) end
		end	
	end
	if bar then
		Nest_SetAttribute(bar, "updated", true) -- for mark/sweep bar deletion
		Nest_SetLabel(bar, label)
		Nest_SetIcon(bar, icon)
		local bc = (vbp.bgColors == "Custom") and vbp.bgColor or c
		local ibr, ibg, ibb, iba = 1, 1, 1, 1
		local ic = vbp.iconBorderColor
		if vbp.iconColors == "Normal" then
			ibr, ibg, ibb, iba = c.r, c.g, c.b, c.a
		elseif (vbp.iconColors == "Debuffs") and (b.barType == "Debuff") then
			local dc = MOD:GetDebuffColorForBar(vbp, b, btype)
			if dc then ibr, ibg, ibb, iba = dc.r, dc.g, dc.b, dc.a end
		elseif (vbp.iconColors == "Custom") and ic then
			ibr, ibg, ibb, iba = ic.r, ic.g, ic.b, ic.a
		end
		ibr, ibg, ibb = Nest_AdjustColor(ibr, ibg, ibb, vbp.borderSaturation or 0, vbp.borderBrightness or 0)
		Nest_SetColors(bar, c.r, c.g, c.b, c.a, bc.r, bc.g, bc.b, bc.a, ibr, ibg, ibb, iba)
		Nest_SetCount(bar, iconCount) -- set the icon text to this count or blank if nil
		Nest_SetAttribute(bar, "tooltipType", ttType)
		Nest_SetAttribute(bar, "tooltipID", ttID)
		Nest_SetAttribute(bar, "tooltipUnit", ttUnit)
		Nest_SetAttribute(bar, "saveBar", b) -- not valid in auto bar group since it then points to a local not a permanent table!
		Nest_SetAttribute(bar, "pulseStart", b.pulseStart) -- pulse icon at start
		Nest_SetAttribute(bar, "pulseEnd", b.pulseEnd) -- pulse icon when expiring
		
		local sbg, sbt, sba = nil, nil, nil
		if b.barType ~= "Notification" then sbg = vbp; sbt = b.barType; sba = b.action end
		Nest_SetAttribute(bar, "saveBarGroup", sbg)
		Nest_SetAttribute(bar, "saveBarType", sbt)
		Nest_SetAttribute(bar, "saveBarAction", sba)
		
		local click, onEnter, onLeave = Bar_OnClick, nil, nil
		if vbp.showTooltips and (vbp.combatTips or not MOD.status.inCombat) then onEnter = Bar_OnEnter; onLeave = Bar_OnLeave end
		Nest_SetCallbacks(bar, click, onEnter, onLeave)
		
		local dft, gd = not vbp.useDefaultFontsAndTextures, MOD.db.global.Defaults -- indicates the bar group has overrides for fonts and textures		
		local faded, alpha = false, 1 -- set flash and adjust alpha for special effects
		if MOD.status.inCombat then alpha = (dft and vbp.combatAlpha or gd.combatAlpha) else alpha = (dft and vbp.alpha or gd.alpha) end
		if (IsOn(b.flashBar) and (b.flashBar == MOD:CheckCondition(b.flashCondition))) or -- conditional flashing
				(Nest_IsTimer(bar) and ((vbp.flashExpiring and vbp.flashTime and (timeLeft < vbp.flashTime)) or -- bar group flash when expiring
				(b.flashExpiring and b.flashTime and (timeLeft < b.flashTime)))) then -- custom bar flash when expiring
			Nest_SetFlash(bar, true)
		elseif IsOn(b.fadeBar) then -- conditional fade for bars is higher priority than bar group setting for delayed fade
			if (b.fadeBar == MOD:CheckCondition(b.fadeCondition)) and b.fadeAlpha then alpha = alpha * b.fadeAlpha; faded = true end
		elseif b.fade and b.fadeAlpha then
			local _, _, _, startTime = Nest_GetTimes(bar) -- get time bar was created
			if ((GetTime() - startTime) > (b.delayTime or 5)) then alpha = alpha * b.fadeAlpha; faded = true end
		elseif vbp.fade and vbp.fadeAlpha then
			local _, _, _, startTime = Nest_GetTimes(bar) -- get time bar was created
			if ((GetTime() - startTime) > (vbp.delayTime or 5)) then alpha = alpha * vbp.fadeAlpha; faded = true end
		else
			Nest_SetFlash(bar, false)
		end
		if not faded then -- fading is highest priority
			if b.startReady then if b.readyAlpha then alpha = alpha * b.readyAlpha end -- this is a ready bar
			elseif b.normalAlpha then alpha = alpha * b.normalAlpha end -- normal bar
		end
		Nest_SetAlpha(bar, alpha)
	end
end

-- Compare caster to enforce "cast by" restrictions
local function CheckCastBy(caster, cb)
	local isMine = (caster == "player")
	local isPet = (caster == "pet")
	local isOther = not (isMine or isPet)
	if not cb then cb = "player" else cb = string.lower(cb) end -- for backward compatibility
	return (((cb == "player") and isMine) or ((cb == "pet") and isPet) or ((cb == "other") and isOther) or (cb == "anyone"))
end

-- Return the source for a buff, debuff or cooldown
local function GetSpellSource(t)
	if not t then return "Unknown" end
	local s = classNames[t.class]
	if not s and t.race then s = "Racial" end
	if not s then s = "Detected" end
	return s
end

-- Check if an action is in the associated filter bar group
local function CheckFilterBarGroup(bgname, btype, action, value)
	if not bgname then return false end
	local bg = MOD.db.profile.BarGroups[bgname]
	if IsOn(bg) then
		if bg.auto then -- auto bar groups look in the filter list (doesn't matter if black list or white list)
			if btype == "Buff" then
				if (bg.filterBuff or bg.showBuff) and bg.filterBuffList and bg.filterBuffList[action] then return true end
			elseif btype == "Debuff" then
				if (bg.filterDebuff or bg.showDebuff) and bg.filterDebuffList and bg.filterDebuffList[action] then return true end
			elseif btype == "Cooldown" then
				if (bg.filterCooldown or bg.showCooldown) and bg.filterCooldownList and bg.filterCooldownList[action] then return true end
			end
		else -- custom bar groups look at the cached info generated from the custom bar list
			local v = GetCache(bg, btype, action)
			if v == value then return true end -- for auras this is which unit is being monitored
		end
	end
	return false
end

-- Check for detected buffs and create bars for them in the specified bar group
-- Detected auras that don't match current bar group settings may need to be added later if the settings change 
local function DetectNewBuffs(unit, n, aura, isBuff, bp, vbp, bg)
	if bp.filterBuff and bp.filterBuffList and bp.filterBuffList[n] then return end -- check if black listed
	if bp.showBuff and not (bp.filterBuffList and bp.filterBuffList[n]) then return end -- check if white listed
	if bp.filterBuffBars and CheckFilterBarGroup(bp.filterBuffBarGroup, "Buff", n, bp.detectBuffsMonitor) then return end -- check if in filter bar group
	local bt = MOD.BuffTable
	if not bt[n] then bt[n] = { det = true } end -- newly detected aura goes into auras table
	local label = MOD:GetLabel(n) -- check if there is a cached label for this action
	local checkTracking = not (bp.detectTracking and bp.detectOnlyTracking)
	if (aura[4] == "Tracking") then checkTracking = bp.detectTracking end
	local isStealable = (aura[7] ~= nil)
	local isMagic = (aura[4] == "Magic")
	local isOther = not isStealable and not isMagic
	local checkTypes = not bp.filterBuffTypes or (bp.detectStealable and isStealable) or (bp.detectMagicBuffs and isMagic) or (bp.detectOtherBuffs and isOther)
	if CheckCastBy(aura[6], bp.detectBuffsCastBy) and CheckTimeAndDuration(bp, aura[2], aura[5]) and checkTracking and checkTypes then
		local b, tag = detectedBar, "Buff"
		b.action = n
		b.barLabel = label
		b.barType = "Buff"
		if aura[6] then tag = tag .. aura[6] elseif aura[10] and (aura[10] > 0) then tag = tag .. tostring(aura[10]) end
		if aura[14] then tag = tag .. tostring(aura[14]) elseif (aura[11] == "weapon") or (aura[11] == "tracking") then tag = tag .. aura[12] end
		b.uniqueID = tag
		b.barSource = GetSpellSource(bt[n])
		UpdateBar(bp, vbp, bg, b, aura[8], aura[2], aura[5], aura[3], aura[4], aura[11], aura[12], unit)
	end
end

-- Check for detected debuffs and create bars for them in the specified bar group 
local function DetectNewDebuffs(unit, n, aura, isBuff, bp, vbp, bg)
	if bp.filterDebuff and bp.filterDebuffList and bp.filterDebuffList[n] then return end -- check if black listed
	if bp.showDebuff and not (bp.filterDebuffList and bp.filterDebuffList[n]) then return end -- check if white listed
	if bp.filterDebuffBars and CheckFilterBarGroup(bp.filterDebuffBarGroup, "Debuff", n, bp.detectDebuffsMonitor) then return end -- check if in filter bar group
	local bt = MOD.DebuffTable
	if not bt[n] then bt[n] = { det = true } end -- newly detected aura goes into auras table
	local label = MOD:GetLabel(n)
	local checkDispel = bp.detectDispellable and MOD:IsDebuffDispellable(n, unit, aura[4])
	local isBoss = (aura[15] ~= nil)
	local isPoison, isCurse, isMagic, isDisease = (aura[4] == "Poison"), (aura[4] == "Curse"), (aura[4] == "Magic"), (aura[4] == "Disease")
	local isOther = not isBoss and not isPoison and not isCurse and not isMagic and not isDisease
	local checkTypes = not bp.filterDebuffTypes or checkDispel or (bp.detectBoss and isBoss) or (bp.detectOtherDebuffs and isOther) or 
		(bp.detectPoison and isPoison) or (bp.detectCurse and isCurse) or (bp.detectMagic and isMagic) or (bp.detectDisease and isDisease)
	if CheckCastBy(aura[6], bp.detectDebuffsCastBy) and CheckTimeAndDuration(bp, aura[2], aura[5]) and checkTypes then
		local b, tag = detectedBar, "Debuff"
		b.action = n
		b.barLabel = label
		b.barType = "Debuff"
		if aura[6] then tag = tag .. aura[6] elseif aura[10] and (aura[10] > 0) then tag = tag .. tostring(aura[10]) end
		if aura[14] then tag = tag .. tostring(aura[14]) elseif (aura[11] == "weapon") or (aura[11] == "tracking") then tag = tag .. aura[12] end
		b.uniqueID = tag
		b.barSource = GetSpellSource(bt[n])
		UpdateBar(bp, vbp, bg, b, aura[8], aura[2], aura[5], aura[3], aura[4], aura[11], aura[12], unit)
	end
end

-- Check if a cooldown is of right type for the specified bar group
local function CheckCooldownType(cd, bp)
	local other, t, s = true, cd[5], cd[6]
	if (t == "spell link") or (t == "spell") then
		other = false
		if bp.detectSpellCooldowns then return true end
	end
	if (t == "inventory") and ((s == "Trinket0Slot") or (s == "Trinket1Slot")) then
		other = false
		if bp.detectTrinketCooldowns then return true end
	end
	if t == "text" then -- might be a potion or elixir
		if (s == "Shared Potion Cooldown") or (s == "Shared Elixir Cooldown") then
			other = false
			if bp.detectPotionCooldowns then return true end
		end
	end
	if other and bp.detectOtherCooldowns then return true end
	return false
end

-- Return true if cooldown is not one of the special case shared ones
local function CheckSharedCooldowns(n, b, bp)
	if bp.detectSharedFrostTraps then
		if n == LSPELL["Freezing Trap"] then return false end
		if n == LSPELL["Ice Trap"] then b.barLabel = L["Frost Traps"] end
	end
	if bp.detectSharedFireTraps then
		if (n == LSPELL["Black Arrow"]) or (n == LSPELL["Immolation Trap"]) then return false end
		if n == LSPELL["Explosive Trap"] then b.barLabel = L["Fire Traps"] end
	end
	if bp.detectSharedShocks then
		if (n == LSPELL["Frost Shock"]) or (n == LSPELL["Flame Shock"]) then return false end
		if n == LSPELL["Earth Shock"] then b.barLabel = L["Shocks"] end
	end
	if bp.detectSharedCrusader then
		if n == LSPELL["Hammer of the Righteous"] then return false end
		if n == LSPELL["Crusader Strike"] then b.barLabel = L["Crusader/Hammer"] end
	end
	return true
end

-- Automatically generate rune cooldown bars for all six rune slots
local runeSlotPrefix = { "(1)  ", "(2)  ", "(5)  ", "(6)  ",  "(3)  ", "(4)  " }
local function AutoRuneBars(bp, vbp, bg)
	if MOD.myClass ~= "DEATHKNIGHT" then return end
	for i = 1, 6 do
		local rune = MOD.runeSlots[i]
		local icon = MOD.runeIcons[rune.rtype]
		local b = detectedBar
		b.action = MOD.runeTypes[rune.rtype]
		b.barLabel = runeSlotPrefix[i] .. b.action
		b.barType = "Cooldown"
		b.uniqueID = "Cooldown"
		b.barSource = "Detected"
		if rune.ready then -- generate ready bar with no duration
			if CheckTimeAndDuration(bp, 0, 0) then
				UpdateBar(bp, vbp, bg, b, icon, 0, 0, nil, nil, "text", b.action, nil)
			end
		else -- generate cooldown timer bar
			local timeLeft = rune.duration - (GetTime() - rune.start)
			if CheckTimeAndDuration(bp, timeLeft, rune.duration) then
				UpdateBar(bp, vbp, bg, b, icon, timeLeft, rune.duration, nil, nil, "text", b.action, nil)
			end
		end
	end
end

-- Check if there are detected cooldowns and conditionally create bars for them in the specified bar group
local function DetectNewCooldowns(n, cd, bp, vbp, bg)
	if bp.filterCooldown and bp.filterCooldownList and bp.filterCooldownList[n] then return end -- check if black listed
	if bp.showCooldown and not (bp.filterCooldownList and bp.filterCooldownList[n]) then return end -- check if white listed
	if bp.filterCooldownBars and CheckFilterBarGroup(bp.filterCooldownBarGroup, "Cooldown", n, true) then return end -- check if in filter bar group
	local cdt = MOD.CooldownTable	
	if not cdt[n] then cdt[n] = { det = true } end -- newly detected cooldown goes into cooldowns table
	local label = MOD:GetLabel(n)	
	if CheckCastBy(cd[7], bp.detectCooldownsBy) and CheckCooldownType(cd, bp) and CheckTimeAndDuration(bp, cd[1], cd[4]) then
		local b = detectedBar
		b.action = n
		b.barLabel = label
		b.barType = "Cooldown"
		b.uniqueID = "Cooldown"
		b.barSource = GetSpellSource(cdt[n])
		if CheckSharedCooldowns(n, b, bp) then
			UpdateBar(bp, vbp, bg, b, cd[2], cd[1], cd[4], nil, nil, cd[5], cd[6], nil)
		end
	end
end

-- Update all bars in bar group (bp), causing them to appear in display bar group (bg) using appearance options (vbp)
-- The show/hide conditions are tested in this function so they are depending on the updating bar group
local function UpdateBarGroupBars(bp, vbp, bg)
	local pst, stat = PartyInfo(), MOD.status	
	local show = (((pst == "solo") and bp.showSolo) or ((pst == "party") and bp.showParty) or ((pst == "raid") and bp.showRaid)) and
		((bp.showCombat and stat.inCombat) or (bp.showOOC and not stat.inCombat)) and (not stat.isResting or bp.showResting) and
		((bp.showBlizz and not MOD.db.profile.hideBlizz) or (bp.showNotBlizz and MOD.db.profile.hideBlizz)) and
		(not stat.isFishing or bp.showFishing) and (not stat.isMounted or bp.showMounted) and 
		(not bp.checkCondition or IsOff(bp.condition) or MOD:CheckCondition(bp.condition)) and
		(not bp.showClasses or not bp.showClasses[MOD.myClass])
	if show then
		if bp.auto then -- if auto bar group then detect new auras and cooldowns
			if bp.detectBuffs then MOD:IterateAuras(bp.detectBuffsMonitor, DetectNewBuffs, true, bp, vbp, bg) end
			if bp.detectDebuffs then MOD:IterateAuras(bp.detectDebuffsMonitor, DetectNewDebuffs, false, bp, vbp, bg) end
			if bp.detectCooldowns then MOD:IterateCooldowns(DetectNewCooldowns, bp, vbp, bg) end
			if bp.detectRuneCooldowns then AutoRuneBars(bp, vbp, bg) end
		else
			for _, bar in pairs(bp.bars) do -- iterate over each bar in the bar group
				if bar.enableBar and (IsOff(bar.hideBar) or (bar.hideBar ~= MOD:CheckCondition(bar.hideCondition))) then
					local t = bar.barType
					if (t == "Buff")  or (t == "Debuff") then
						local aname, cb, saveLabel, count = bar.action, string.lower(bar.castBy), bar.barLabel, 0
						local auraList = MOD:CheckAura(bar.monitor, aname, t == "Buff")
						if #auraList > 0 then
							for _, aura in pairs(auraList) do
								local isMine, isPet = (aura[6] == "player"), (aura[6] == "pet") -- enforce optional castBy restrictions
								local mon = ((cb == "player") and isMine) or ((cb == "pet") and isPet) or ((cb == "other") and not isMine) or (cb == "anyone")
								if mon and CheckTimeAndDuration(bp, aura[2], aura[5]) then
									count = count + 1
									if count > 1 then bar.barLabel = bar.barLabel .. " " end -- add space at end to make unique
									UpdateBar(bp, vbp, bg, bar, aura[8], aura[2], aura[5], aura[3], aura[4], aura[11], aura[12], bar.monitor)
									bar.startReady = nil
								end
							end
							bar.barLabel = saveLabel -- restore in case of multiple bar copies
						elseif bar.enableReady then -- see if need to create a ready bar for spell off cooldown
							if not bar.readyTime or (bar.readyTime == 0) then bar.startReady = nil end
							if not bar.startReady or ((GetTime() - bar.startReady) < bar.readyTime) then
								if not bar.startReady then bar.startReady = GetTime() end
								local bt = (t == "Buff") and MOD.BuffTable or MOD.DebuffTable
								local aura, ttype, link = bt[aname]
								if aura then link = GetSpellLink(aura.id or aname) end
								if link then ttype = "spell link" else ttype = "text"; link = aname end
								UpdateBar(bp, vbp, bg, bar, MOD:GetIcon(aname), 0, 0, nil, nil, ttype, link, nil)
							end
						end
					elseif t == "Cooldown" then
						local aname = bar.action
						local cd = MOD:CheckCooldown(aname) -- look up in the active cooldowns table
						if cd and (cd[1] ~= nil) then
							if CheckTimeAndDuration(bp, cd[1], cd[4]) then
								UpdateBar(bp, vbp, bg, bar, cd[2], cd[1], cd[4], nil, nil, cd[5], cd[6], nil)
								bar.startReady = nil
							end
						elseif bar.enableReady and (bar.readyNotUsable or IsUsableSpell(aname)) then -- see if need to create a ready bar
							if not bar.readyTime or (bar.readyTime == 0) then bar.startReady = nil end
							if not bar.startReady or ((GetTime() - bar.startReady) < bar.readyTime) then
								if not bar.startReady then bar.startReady = GetTime() end
								local ttype, link = "spell link", MOD:GetHyperlink(aname)
								if not link then ttype = "text"; link = aname end
								UpdateBar(bp, vbp, bg, bar, MOD:GetIcon(aname), 0, 0, nil, nil, ttype, link, nil)
							end
						end
					elseif t == "Notification" then
						if MOD:CheckCondition(bar.action) then
							local icon = MOD:GetIconForBar(bar)
							if not icon then icon = defaultNotificationIcon end
							UpdateBar(bp, vbp, bg, bar, icon, 0, 0, nil, nil, "notification", bar.barLabel, bar.action)
						end
					end
				end
			end
		end
	end
end

-- Update bars in all bar groups, checking for bar group visibility and removing expired bars
function MOD:UpdateBars()
	if hidden then -- if was hidden then need to re-initialize all bar groups
		hidden = false
		MOD:UpdateAllBarGroups()
	end
	
	for _, bp in pairs(MOD.db.profile.BarGroups) do -- iterate through the all bar groups
		if IsOn(bp) then
			local bg = Nest_GetBarGroup(bp.name) -- match the profile bar group to the LibBars bar group
			if bg then	
				if bp.enabled then -- check all the conditions under which the bar group might be hidden are not true
					Nest_SetAllAttributes(bg, "updated", false) -- first, mark all the bars in the group as not updated...
					if not bp.merged then
						UpdateBarGroupBars(bp, bp, bg) -- then update all the bars for the bar group into the display bar group					
						for _, mbp in pairs(MOD.db.profile.BarGroups) do -- then look for bar groups merging into this bar group
							if IsOn(mbp) and mbp.enabled and mbp.merged and (mbp.mergeInto == bp.name) then
								UpdateBarGroupBars(mbp, bp, bg) -- update all bars for merged bar group into same display bar group
							end
						end
					end				
					UpdateTestBars(bg) -- update any unexpired test bars in this bar group
					Nest_DeleteBarsWithAttribute(bg, "updated", false) -- then, remove any bars in the group that weren't updated
				else -- if not then hide any bars that might be around
					Nest_DeleteAllBars(bg)
				end
			end
		end
	end
	UpdatePositions()
end
